import dis
import queue
import sys
import threading
import traceback
import unittest
from collections import Counter
from unittest.mock import Mock, call
from numba.tests.support import TestCase
from numba import jit, objmode
from numba.core.utils import PYVERSION
from numba.core.serialize import _numba_unpickle


def _enable_sysmon(disp):
    """Decorator to enable sys.monitoring on the dispatcher"""
    disp._enable_sysmon = True
    return disp


def generate_usecase():
    @_enable_sysmon
    @jit('int64(int64)',)
    def foo(x):
        return x + 1

    def call_foo(x):
        return 2 * foo(x + 5)

    return foo, call_foo


if PYVERSION in ((3, 12), (3, 13), (3, 14)):
    PY_START = sys.monitoring.events.PY_START
    PY_RETURN = sys.monitoring.events.PY_RETURN
    RAISE = sys.monitoring.events.RAISE
    PY_UNWIND = sys.monitoring.events.PY_UNWIND
    STOP_ITERATION = sys.monitoring.events.STOP_ITERATION
    NO_EVENTS = sys.monitoring.events.NO_EVENTS


TOOL2MONITORTYPE = {0 : "Debugger",
                    1 : "Coverage",
                    2 : "Profiler",
                    5 : "Optimizer"}


@unittest.skipUnless(PYVERSION >= (3, 12), "needs Python 3.12+")
class TestMonitoring(TestCase):
    # Tests the interaction of the Numba dispatcher with `sys.monitoring`.
    #
    # Note that it looks like a lot of these try..finally type patterns could
    # be written using a context manager, this is true, but it is not written
    # like that deliberately as a context manager adds implementation details
    # onto the stack which makes it harder to debug tests.

    def setUp(self):
        # First... check if there's other monitoring stuff registered (e.g. test
        # is running under cProfile or coverage), skip if so.
        monitor_kinds = []
        for i in range(6): # there are 5 tool IDs
            if sys.monitoring.get_tool(i) is not None:
                monitor_kinds.append(TOOL2MONITORTYPE[i])

        if monitor_kinds:
            msg = ("Cannot run monitoring tests when other monitors are "
                   "active, found monitor(s) of type: "
                   f"{', '.join(monitor_kinds)}")
            self.skipTest(msg)

        # set up some standard functions and answers for use throughout
        self.foo, self.call_foo = generate_usecase()
        self.arg = 10
        self.foo_result = self.arg + 5 + 1
        self.call_foo_result = 2 * self.foo_result
        # pretend to be a profiler in the majority of these unit tests
        self.tool_id = sys.monitoring.PROFILER_ID

    def gather_mock_calls_multithreads(self, mockcalls):
        # Gather mock-calls for the `self.foo` and `self.call_foo`
        matched = Counter()
        target_codeobjs = {self.call_foo.__code__, self.foo.__code__}
        for cb_args in mockcalls._mock_call_args_list:
            (codeobj, *args) = cb_args.args
            if codeobj in target_codeobjs:
                matched[(codeobj, *args)] += 1
        return matched

    def check_py_start_calls_multithreads(self, allcalls):
        # Checks that PY_START calls were correctly captured for a
        # `self.call_foo(self.arg)` call in multithreads
        matched = self.gather_mock_calls_multithreads(allcalls[PY_START])
        self.assertEqual(len(matched), 2) # two types of call

        # Find the resume op, this is where the code for `call_foo` "starts"
        inst = [x for x in dis.get_instructions(self.call_foo)
                if x.opname == "RESUME"]
        offset = inst[0].offset
        self.assertEqual(matched[self.call_foo.__code__, offset], 2)
        self.assertEqual(matched[self.foo.__code__, 0], 2)
        self.assertEqual(matched.total(), 4)

    def check_py_start_calls(self, allcalls):
        # Checks that PY_START calls were correctly captured for a
        # `self.call_foo(self.arg)` call.
        mockcalls = allcalls[PY_START]
        self.assertEqual(mockcalls.call_count, 2)
        # Find the resume op, this is where the code for `call_foo` "starts"
        inst = [x for x in dis.get_instructions(self.call_foo)
                if x.opname == "RESUME"]
        offset = inst[0].offset
        # Numba always reports the start location as offset 0.
        calls = (call(self.call_foo.__code__, offset),
                 call(self.foo.__code__, 0))
        mockcalls.assert_has_calls(calls)

    def check_py_return_calls_multithreads(self, allcalls):
        # Checks that PY_RETURN calls were correctly captured for a
        # `self.call_foo(self.arg)` call.
        matched = self.gather_mock_calls_multithreads(allcalls[PY_RETURN])
        offset = [x for x in dis.get_instructions(self.call_foo)][-1].offset
        self.assertEqual(matched[self.foo.__code__, 0, self.foo_result], 2)
        self.assertEqual(
            matched[self.call_foo.__code__, offset, self.call_foo_result], 2
        )
        self.assertEqual(matched.total(), 4)

    def check_py_return_calls(self, allcalls):
        # Checks that PY_RETURN calls were correctly captured for a
        # `self.call_foo(self.arg)` call.
        mockcalls = allcalls[PY_RETURN]
        self.assertEqual(mockcalls.call_count, 2)
        # These are in the order the returns were encountered. Return from `foo`
        # occurred first, followed by return from `call_foo`.
        # NOTE: it is a known issue that Numba reports the PY_RETURN event as
        # occurring at offset 0. At present there's no information about the
        # location that the return occurred propagating from the machine code
        # back to the dispatcher (where the monitoring events are handled).
        offset = [x for x in dis.get_instructions(self.call_foo)][-1].offset
        calls = [call(self.foo.__code__, 0, self.foo_result),
                 call(self.call_foo.__code__, offset, self.call_foo_result)]
        mockcalls.assert_has_calls(calls)

    def run_with_events(self, function, args, events, tool_id=None,
                        barrier=None):
        # Runs function with args with monitoring set for events on `tool_id`
        # (if present, else just uses the default of "PROFILER_ID") returns a
        # dictionary event->callback.
        try:
            if tool_id is None:
                _tool_id = self.tool_id
            else:
                _tool_id = tool_id
            sys.monitoring.use_tool_id(_tool_id, "custom_monitor")
            callbacks = {}
            event_bitmask = 0
            for event in events:
                callback = Mock()
                sys.monitoring.register_callback(_tool_id, event, callback)
                callbacks[event] = callback
                event_bitmask |= event
            # only start monitoring once callbacks are registered
            sys.monitoring.set_events(_tool_id, event_bitmask)
            if barrier is not None:
                # Wait for all threads to have enabled events.
                barrier()
            function(*args)
        finally:
            # clean up state
            if barrier is not None:
                # Wait for all threads to finish `function()`
                # This makes sure all threads have a chance to see the events
                # from other threads.
                barrier()
            sys.monitoring.set_events(_tool_id, NO_EVENTS)
            for event in events:
                sys.monitoring.register_callback(_tool_id, event, None)
            sys.monitoring.free_tool_id(_tool_id)
        return callbacks

    def test_start_event(self):
        # test event PY_START
        cb = self.run_with_events(self.call_foo, (self.arg,), (PY_START,))
        # Check...
        self.assertEqual(len(cb), 1)
        self.check_py_start_calls(cb)

    def test_return_event(self):
        # test event PY_RETURN
        cb = self.run_with_events(self.call_foo, (self.arg,), (PY_RETURN,))
        # Check...
        self.assertEqual(len(cb), 1)
        self.check_py_return_calls(cb)

    def test_call_event_chain(self):
        # test event PY_START and PY_RETURN monitored at the same time
        cb = self.run_with_events(self.call_foo, (self.arg,),
                                  (PY_START, PY_RETURN))
        # Check...
        self.assertEqual(len(cb), 2)
        self.check_py_return_calls(cb)
        self.check_py_start_calls(cb)

    # --------------------------------------------------------------------------
    # NOTE: About the next two tests...
    # Numba doesn't support "local event" level monitoring, it's implemented
    # in CPython via adjusting the code object bytecode to use
    # "instrumented" opcodes. When the interpreter encounters an
    # instrumented opcode it triggers the event handling pathways. As Numba
    # doesn't interpret the bytecode instruction-at-a-time there's not
    # really any way to support this. Two things to check...
    # 1. The an instrumented code object doesn't trigger events in
    #    the dispatcher.
    # 2. That Numba can compile instrumented functions (it should be able
    #    to without any problem as the instrumented bytecode should not
    #    leak into `.co_code`.).

    def test_instrumented_code_does_not_trigger_numba_events(self):
        # 1. from above.
        @jit('int64(int64)',)
        def foo(x):
            return x + 3

        try:
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            callbacks = {}
            event_bitmask = 0
            events = (PY_START, PY_RETURN)
            for event in events:
                callback = Mock()
                sys.monitoring.register_callback(tool_id, event, callback)
                callbacks[event] = callback
                event_bitmask |= event

            sys.monitoring.set_local_events(tool_id, foo.__code__,
                                            event_bitmask)
            result = foo(self.arg)
        finally:
            for event in events:
                sys.monitoring.register_callback(tool_id, event, None)
            sys.monitoring.set_local_events(tool_id, foo.__code__, 0)
            sys.monitoring.free_tool_id(tool_id)

        # check
        self.assertEqual(result, foo.py_func(self.arg))
        self.assertEqual(len(callbacks), 2)
        callbacks[PY_START].assert_not_called()
        callbacks[PY_RETURN].assert_not_called()

    def test_instrumented_code_can_be_compiled(self):
        # 2. from above.

        def foo(x):
            return x + 1

        try:
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.set_local_events(tool_id, foo.__code__, PY_START)
            sys.monitoring.register_callback(tool_id, PY_START, Mock())
            # test compile
            result = jit(foo)(self.arg)
            self.assertEqual(result, foo(self.arg))
        finally:
            sys.monitoring.register_callback(tool_id, PY_START, None)
            sys.monitoring.set_local_events(tool_id, foo.__code__, 0)
            sys.monitoring.free_tool_id(tool_id)

    def test_unhandled_events_are_ignored(self):
        # Check an unhandled event e.g. PY_YIELD isn't reported.
        def generate(dec):
            @dec('void()')
            def producer():
                yield 10

            @dec('int64()')
            def consumer():
                p = producer()
                return next(p)

            return consumer

        event = sys.monitoring.events.PY_YIELD
        # check that pure python reports
        wrapper = lambda sig: lambda fn: fn
        py_consumer = generate(wrapper)
        py_cb = self.run_with_events(py_consumer, (),  (event,))
        py_cb[event].assert_called_once()
        # check the numba does not report
        nb_consumer = generate(jit)
        nb_cb = self.run_with_events(nb_consumer, (),  (event,))
        nb_cb[event].assert_not_called()

    def test_event_with_no_callback_runs(self):
        # This checks the situation where an event is being monitored but
        # there's no callback associated with the event. In the dispatcher C
        # code the loop over tools will be entered, but nothing will get called
        # as the "instrument" is missing (NULL).
        try:
            event = PY_START
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.set_events(tool_id, event)
            # NO CALLBACK IS REGISTERED!
            active_events = sys.monitoring.get_events(tool_id)
            self.assertEqual(active_events, event)
            result = self.call_foo(self.arg)
            active_events = sys.monitoring.get_events(tool_id)
            self.assertEqual(active_events, event)
            self.assertEqual(result, self.call_foo_result)
        finally:
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.free_tool_id(tool_id)

    def test_disable_from_callback(self):
        # Event callbacks can disable a _local_ event at a specific location to
        # prevent it triggering in the future by returning
        # `sys.monitoring.DISABLE`. As this only applies to local events, doing
        # this should have absolutely no impact for the global events that Numba
        # supports.

        callback = Mock(return_value=sys.monitoring.DISABLE)

        try:
            event = PY_START
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.set_events(tool_id, event)
            sys.monitoring.register_callback(tool_id, event, callback)
            active_events = sys.monitoring.get_events(tool_id)
            self.assertEqual(active_events, event)
            result = self.call_foo(self.arg)
            active_events = sys.monitoring.get_events(tool_id)
            self.assertEqual(active_events, event)
            self.assertEqual(result, self.call_foo_result)
            callback.assert_called()
        finally:
            # It is necessary to restart events that have been disabled. The
            # "disabled" state of the `PY_START` event for the tool
            # `self.tool_id` "leaks" into subsequent tests. These subsequent
            # tests then end up failing as events that should have been
            # triggered are not triggered due to the state leak! It's not really
            # clear why this happens, if it is part of the design or a side
            # effect of the design, or if this behaviour is simply a bug in
            # CPython itself.
            sys.monitoring.restart_events()
            sys.monitoring.register_callback(tool_id, event, None)
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.free_tool_id(tool_id)

    def test_mutation_from_objmode(self):
        try:
            # Check that it's possible to enable an event (mutate the event
            # state)from an `objmode` block. Monitoring for PY_RETURN is set in
            # objmode once the function starts executing.
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            event = PY_RETURN
            # register the callback... note that the event isn't switched on yet
            callback = Mock()
            sys.monitoring.register_callback(tool_id, event, callback)

            def objmode_enable_event(switch_on_event):
                if switch_on_event:
                    sys.monitoring.set_events(tool_id, event)

            @_enable_sysmon
            @jit('int64(int64)')
            def foo(enable):
                with objmode:
                    objmode_enable_event(enable)
                return enable + 7

            # this should not trigger the return callback
            foo(0)
            callback.assert_not_called()

            # this should trigger the return callback
            foo(1)
            # switch off the event so the callback mock is protected from
            # mutation.
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            # check what happened
            callback.assert_called()
            # 2 calls, 1 is the return from the objmode_enable_event, the other
            # is the return from foo.
            self.assertEqual(callback.call_count, 2)
        finally:
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.register_callback(tool_id, event, None)
            sys.monitoring.free_tool_id(tool_id)

    def test_multiple_tool_id(self):
        # Check that multiple tools will work across different combinations of
        # events that Numba dispatcher supports, namely:
        # (NO_EVENTS, PY_START, PY_RETURN).

        # the use of NO_EVENTS is superfluous, it is to demonstrate usage.
        tool_ids_2_events = {sys.monitoring.DEBUGGER_ID: (NO_EVENTS,),
                             sys.monitoring.COVERAGE_ID: (PY_START,),
                             sys.monitoring.PROFILER_ID: (PY_RETURN,),
                             sys.monitoring.OPTIMIZER_ID:
                                 (PY_START, PY_RETURN,),}

        all_callbacks = {}
        try:
            for tool_id, events in tool_ids_2_events.items():
                sys.monitoring.use_tool_id(tool_id, f"custom_monitor_{tool_id}")
                event_bitmask = 0
                callbacks = {}
                all_callbacks[tool_id] = callbacks
                for event in events:
                    callback = Mock()
                    # Can't set an event for NO_EVENTS!
                    if event != NO_EVENTS:
                        sys.monitoring.register_callback(tool_id, event,
                                                         callback)
                    callbacks[event] = callback
                    event_bitmask |= event
                # only start monitoring once callbacks are registered
            for tool_id in tool_ids_2_events.keys():
                sys.monitoring.set_events(tool_id, event_bitmask)
            self.call_foo(self.arg)
        finally:
            # clean up state
            for tool_id, events in tool_ids_2_events.items():
                for event in events:
                    # Can't remove an event for NO_EVENTS!
                    if event != NO_EVENTS:
                        sys.monitoring.register_callback(tool_id, event, None)
                sys.monitoring.set_events(tool_id, NO_EVENTS)
                sys.monitoring.free_tool_id(tool_id)

        # Now check all_callbacks...

        # check debugger tool slot
        dbg_tool = all_callbacks[sys.monitoring.DEBUGGER_ID]
        self.assertEqual(len(dbg_tool), 1) # one event to capture
        callback = dbg_tool[NO_EVENTS]
        callback.assert_not_called()

        # check coverage tool slot
        cov_tool = all_callbacks[sys.monitoring.COVERAGE_ID]
        self.assertEqual(len(cov_tool), 1) # one event to capture
        self.check_py_start_calls(cov_tool)

        # check profiler tool slot
        prof_tool = all_callbacks[sys.monitoring.PROFILER_ID]
        self.assertEqual(len(prof_tool), 1) # one event to capture
        self.check_py_return_calls(prof_tool)

        # check optimiser tool slot
        opt_tool = all_callbacks[sys.monitoring.OPTIMIZER_ID]
        self.assertEqual(len(opt_tool), 2) # two events to capture
        self.check_py_start_calls(opt_tool)
        self.check_py_return_calls(opt_tool)

    def test_raising_under_monitoring(self):
        # Check that Numba can raise an exception whilst monitoring is running
        # and that 1. `RAISE` is issued 2. `PY_UNWIND` is issued, 3. that
        # `PY_RETURN` is not issued.

        ret_callback = Mock()
        raise_callback = Mock()
        unwind_callback = Mock()

        msg = 'exception raised'

        @_enable_sysmon
        @jit('()')
        def foo():
            raise ValueError(msg)

        store_raised = None
        try:
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.register_callback(tool_id, PY_RETURN, ret_callback)
            sys.monitoring.register_callback(tool_id, RAISE, raise_callback)
            sys.monitoring.register_callback(tool_id, PY_UNWIND,
                                             unwind_callback)
            sys.monitoring.set_events(tool_id, PY_RETURN | RAISE | PY_UNWIND)
            try:
                foo()
            except ValueError as raises:
                store_raised = raises
            # switch off monitoring
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            # check that the ret_callback was called once (by Numba unpickle to
            # fetch the exception info out of the stored bytes).
            ret_callback.assert_called_once()
            # and that elements that are feasible to check about the call are
            # as expected
            the_call = ret_callback.call_args_list[0]
            self.assertEqual(the_call.args[0], _numba_unpickle.__code__)
            self.assertEqual(the_call.args[2][0], ValueError)
            self.assertEqual(the_call.args[2][1][0], msg)

            # check that the RAISE event callback was triggered
            raise_callback.assert_called()
            numba_unpickle_call = raise_callback.call_args_list[0]
            self.assertEqual(numba_unpickle_call.args[0],
                             _numba_unpickle.__code__)
            self.assertIsInstance(numba_unpickle_call.args[2], KeyError)
            foo_call = raise_callback.call_args_list[1]
            self.assertEqual(foo_call.args[0], foo.py_func.__code__)
            self.assertIsInstance(foo_call.args[2], ValueError)
            self.assertIn(msg, str(foo_call.args[2]))

            # check that PY_UNWIND event callback was called
            unwind_callback.assert_called_once()
            unwind_call = unwind_callback.call_args_list[0]
            self.assertEqual(unwind_call.args[0], foo.py_func.__code__)
            self.assertIsInstance(unwind_call.args[2], ValueError)
            self.assertIn(msg, str(unwind_call.args[2]))
        finally:
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.register_callback(tool_id, PY_RETURN, None)
            sys.monitoring.register_callback(tool_id, RAISE, None)
            sys.monitoring.register_callback(tool_id, PY_UNWIND, None)
            sys.monitoring.free_tool_id(tool_id)

        self.assertIn(msg, str(store_raised))

    def test_stop_iteration_under_monitoring(self):
        # Check that Numba can raise an StopIteration exception whilst
        # monitoring is running and that:
        # 1. RAISE is issued for an explicitly raised StopIteration exception.
        # 2. PY_RETURN is issued appropriately for the unwinding stack
        # 3. STOP_ITERATION is not issued as there is no implicit StopIteration
        #    raised.

        return_callback = Mock()
        raise_callback = Mock()
        stopiter_callback = Mock()

        msg = 'exception raised'

        @_enable_sysmon
        @jit('()')
        def foo():
            raise StopIteration(msg)

        store_raised = None
        try:
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.register_callback(tool_id, PY_RETURN,
                                             return_callback)
            sys.monitoring.register_callback(tool_id, RAISE,
                                             raise_callback)
            sys.monitoring.register_callback(tool_id, STOP_ITERATION,
                                             stopiter_callback)
            sys.monitoring.set_events(tool_id,
                                      PY_RETURN | STOP_ITERATION | RAISE)
            try:
                foo()
            except StopIteration as raises:
                store_raised = raises
            # switch off monitoring
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            # check that the return_callback was called once (by Numba unpickle
            # to fetch the exception info out of the stored bytes).
            return_callback.assert_called_once()
            # and that elements that are feasible to check about the call are
            # as expected
            the_call = return_callback.call_args_list[0]
            self.assertEqual(the_call.args[0], _numba_unpickle.__code__)
            self.assertEqual(the_call.args[2][0], StopIteration)
            self.assertEqual(the_call.args[2][1][0], msg)

            # check that the RAISE event callback was triggered
            raise_callback.assert_called()
            # check that it's 3 long (numba unpickle, jit(foo), the test method)
            self.assertEqual(raise_callback.call_count, 3)

            # check the numba pickle call
            numba_unpickle_call = raise_callback.call_args_list[0]
            self.assertEqual(numba_unpickle_call.args[0],
                             _numba_unpickle.__code__)
            self.assertIsInstance(numba_unpickle_call.args[2], KeyError)

            # check the jit(foo) call
            foo_call = raise_callback.call_args_list[1]
            self.assertEqual(foo_call.args[0], foo.py_func.__code__)
            self.assertIsInstance(foo_call.args[2], StopIteration)
            self.assertIn(msg, str(foo_call.args[2]))

            # check the test method call
            meth_call = raise_callback.call_args_list[2]
            test_method_code = sys._getframe().f_code
            self.assertEqual(meth_call.args[0], test_method_code)
            self.assertIsInstance(meth_call.args[2], StopIteration)
            self.assertIn(msg, str(meth_call.args[2]))

            # check that the STOP_ITERATION event was not triggered
            stopiter_callback.assert_not_called()
        finally:
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.register_callback(tool_id, PY_RETURN, None)
            sys.monitoring.register_callback(tool_id, STOP_ITERATION, None)
            sys.monitoring.register_callback(tool_id, RAISE, None)
            sys.monitoring.free_tool_id(tool_id)

        self.assertIn(msg, str(store_raised))

    def test_raising_callback_unwinds_from_jit_on_success_path(self):
        # An event callback can legitimately raise an exception, this test
        # makes sure Numba's dispatcher handles it ok on the "successful path",
        # i.e. the JIT compiled function didn't raise an exception at runtime.

        msg = "deliberately broken callback"

        callback = Mock(side_effect=ValueError(msg))

        store_raised = None
        try:
            event = PY_START
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.set_events(tool_id, event)
            sys.monitoring.register_callback(tool_id, event, callback)
            self.foo(self.arg)
        except ValueError as raises:
            store_raised = raises
        finally:
            sys.monitoring.register_callback(tool_id, event, None)
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.free_tool_id(tool_id)

        callback.assert_called_once()
        self.assertIn(msg, str(store_raised))

    def test_raising_callback_unwinds_from_jit_on_raising_path(self):
        # An event callback can legitimately raise an exception, this test
        # makes sure Numba's dispatcher handles it ok on the
        # "unsuccessful path", i.e. the JIT compiled function raised an
        # exception at runtime. This test checks the RAISE event, as the
        # callback itself raises, it overrides the exception coming from the
        # JIT compiled function.

        msg_callback = "deliberately broken callback"
        msg_execution = "deliberately broken execution"

        callback = Mock(side_effect=ValueError(msg_callback))

        class LocalException(Exception):
            pass

        @_enable_sysmon
        @jit("()")
        def raising():
            raise LocalException(msg_execution)

        store_raised = None
        try:
            event = RAISE
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.set_events(tool_id, event)
            sys.monitoring.register_callback(tool_id, event, callback)
            raising()
        except ValueError as raises:
            store_raised = raises
        finally:
            sys.monitoring.register_callback(tool_id, event, None)
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.free_tool_id(tool_id)

        callback.assert_called()
        # Called 3x (numba unpickle, ValueError in callback, the test method)
        self.assertEqual(callback.call_count, 3)

        # check the numba unpickle call
        numba_unpickle_call = callback.call_args_list[0]
        self.assertEqual(numba_unpickle_call.args[0], _numba_unpickle.__code__)
        self.assertIsInstance(numba_unpickle_call.args[2], KeyError)

        # check the jit(raising) call
        raising_call = callback.call_args_list[1]
        self.assertEqual(raising_call.args[0], raising.py_func.__code__)
        self.assertIs(raising_call.args[2], callback.side_effect)

        # check the test method call
        meth_call = callback.call_args_list[2]
        test_method_code = sys._getframe().f_code
        self.assertEqual(meth_call.args[0], test_method_code)
        self.assertIs(meth_call.args[2], callback.side_effect)

        # check the stored exception is the expected exception
        self.assertIs(store_raised, callback.side_effect)

    def test_raising_callback_unwinds_from_jit_on_unwind_path(self):
        # An event callback can legitimately raise an exception, this test
        # makes sure Numba's dispatcher handles it ok on the
        # "unsuccessful path", i.e. the JIT compiled function raised an
        # exception at runtime. This test checks the PY_UNWIND event. CPython
        # seems to not notice the PY_UNWIND coming from the exception arising
        # from the raise in the event callback, it just has the PY_UNWIND from
        # the raise in the JIT compiled function.

        msg_callback = "deliberately broken callback"
        msg_execution = "deliberately broken execution"

        callback = Mock(side_effect=ValueError(msg_callback))

        class LocalException(Exception):
            pass

        @_enable_sysmon
        @jit("()")
        def raising():
            raise LocalException(msg_execution)

        store_raised = None
        try:
            event = PY_UNWIND
            tool_id = self.tool_id
            sys.monitoring.use_tool_id(tool_id, "custom_monitor")
            sys.monitoring.set_events(tool_id, event)
            sys.monitoring.register_callback(tool_id, event, callback)
            raising()
        except ValueError as raises:
            store_raised = raises
        finally:
            sys.monitoring.register_callback(tool_id, event, None)
            sys.monitoring.set_events(tool_id, NO_EVENTS)
            sys.monitoring.free_tool_id(tool_id)

        callback.assert_called_once()

        # check the jit(raising) call
        raising_call = callback.call_args_list[0]
        self.assertEqual(raising_call.args[0], raising.py_func.__code__)
        self.assertEqual(type(raising_call.args[2]), LocalException)
        self.assertEqual(str(raising_call.args[2]), msg_execution)

        # check the stored_raise
        self.assertIs(store_raised, callback.side_effect)

    def test_monitoring_multiple_threads(self):
        # Two threads, different tools and events registered on each thread.
        # Each test creates a global event capturing. The threads use barriers
        # to wait for each other to start and stop capturing. This way they
        # see the events from each other. One thread is capturing PY_START
        # and the other is capturing PY_RETURN.
        barrier = threading.Barrier(2)

        def barrier_cb():
            barrier.wait()

        def t1_work(self, q):
            try:
                # test event PY_START on a "debugger tool"
                cb = self.run_with_events(self.call_foo, (self.arg,),
                                          (PY_START,),
                                          tool_id=sys.monitoring.DEBUGGER_ID,
                                          barrier=barrier_cb)
                # Check...
                self.assertEqual(len(cb), 1)
                self.check_py_start_calls_multithreads(cb)
            except Exception as e:
                q.put(''.join(traceback.format_exception(e)))

        def t2_work(self, q):
            try:
                # test event PY_RETURN on a "coverage tool"
                cb = self.run_with_events(self.call_foo, (self.arg,),
                                          (PY_RETURN,),
                                          tool_id=sys.monitoring.COVERAGE_ID,
                                          barrier=barrier_cb)
                # Check...
                self.assertEqual(len(cb), 1)
                self.check_py_return_calls_multithreads(cb)
            except Exception as e:
                q.put(''.join(traceback.format_exception(e)))

        q1 = queue.Queue()
        t1 = threading.Thread(target=t1_work, args=(self, q1))
        q2 = queue.Queue()
        t2 = threading.Thread(target=t2_work, args=(self, q2))

        threads = (t1, t2)
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        # make sure there were no exceptions
        def assert_empty_queue(q):
            if q.qsize() != 0:
                while not q.empty():
                    print(q.get())
                self.fail("queue supposed to be empty")

        assert_empty_queue(q1)
        assert_empty_queue(q2)


@unittest.skipUnless(PYVERSION >= (3, 12), "needs Python 3.12+")
class TestMonitoringSelfTest(TestCase):

    def test_skipping_of_tests_if_monitoring_in_use(self):
        # check that the unit tests in the TestMonitoring class above will skip
        # if there are other monitoring tools registered in the thread (in this
        # case cProfile is used to cause that effect).
        r = self.subprocess_test_runner(TestMonitoring.__module__,
                                        'TestMonitoring',
                                        'test_start_event',
                                        flags={'-m': 'cProfile'})
        self.assertIn("skipped=1", str(r))


@unittest.skipUnless(PYVERSION >= (3, 12), "needs Python 3.12+")
class TestMonitoringEnvVarControl(TestCase):
    @TestCase.run_test_in_subprocess(
        envvars={"NUMBA_ENABLE_SYS_MONITORING": ''})
    def test_default_off(self):
        @jit
        def foo(x):
            return x + 1

        self.assertFalse(foo._enable_sysmon)

    @TestCase.run_test_in_subprocess(
        envvars={"NUMBA_ENABLE_SYS_MONITORING": '0'})
    def test_override_off(self):
        @jit
        def foo(x):
            return x + 1

        self.assertFalse(foo._enable_sysmon)

    @TestCase.run_test_in_subprocess(
        envvars={"NUMBA_ENABLE_SYS_MONITORING": '1'})
    def test_override_on(self):
        @jit
        def foo(x):
            return x + 1

        self.assertTrue(foo._enable_sysmon)


if __name__ == '__main__':
    unittest.main()



import collections

import numpy as np

from numba.core import types, config


QuicksortImplementation = collections.namedtuple(
    'QuicksortImplementation',
    (# The compile function itself
     'compile',
     # All subroutines exercised by test_sort
     'partition', 'partition3', 'insertion_sort',
     # The top-level function
     'run_quicksort',
     ))


Partition = collections.namedtuple('Partition', ('start', 'stop'))

# Under this size, switch to a simple insertion sort
SMALL_QUICKSORT = 15

MAX_STACK = 100


def make_quicksort_impl(wrap, lt=None, is_argsort=False, is_list=False, is_np_array=False):

    if config.USE_LEGACY_TYPE_SYSTEM:
        intp = types.intp
    else:
        intp = types.py_int
    zero = intp(0)

    # Two subroutines to make the core algorithm generic wrt. argsort
    # or normal sorting.  Note the genericity may make basic sort()
    # slightly slower (~5%)
    if is_argsort:
        if is_list:
            @wrap
            def make_res(A):
                return [x for x in range(len(A))]
        else:
            @wrap
            def make_res(A):
                return np.arange(A.size)

        @wrap
        def GET(A, idx_or_val):
            return A[idx_or_val]

    else:
        @wrap
        def make_res(A):
            return A

        @wrap
        def GET(A, idx_or_val):
            return idx_or_val

    def default_lt(a, b):
        """
        Trivial comparison function between two keys.
        """
        return a < b

    LT = wrap(lt if lt is not None else default_lt)

    @wrap
    def insertion_sort(A, R, low, high):
        """
        Insertion sort A[low:high + 1]. Note the inclusive bounds.
        """
        assert low >= 0
        if high <= low:
            return

        for i in range(low + 1, high + 1):
            k = R[i]
            v = GET(A, k)
            # Insert v into A[low:i]
            j = i
            while j > low and LT(v, GET(A, R[j - 1])):
                # Make place for moving A[i] downwards
                R[j] = R[j - 1]
                j -= 1
            R[j] = k

    @wrap
    def partition(A, R, low, high):
        """
        Partition A[low:high + 1] around a chosen pivot.  The pivot's index
        is returned.
        """
        assert low >= 0
        assert high > low

        mid = (low + high) >> 1
        # NOTE: the pattern of swaps below for the pivot choice and the
        # partitioning gives good results (i.e. regular O(n log n))
        # on sorted, reverse-sorted, and uniform arrays.  Subtle changes
        # risk breaking this property.

        # median of three {low, middle, high}
        if LT(GET(A, R[mid]), GET(A, R[low])):
            R[low], R[mid] = R[mid], R[low]
        if LT(GET(A, R[high]), GET(A, R[mid])):
            R[high], R[mid] = R[mid], R[high]
        if LT(GET(A, R[mid]), GET(A, R[low])):
            R[low], R[mid] = R[mid], R[low]
        pivot = GET(A, R[mid])

        # Temporarily stash the pivot at the end
        R[high], R[mid] = R[mid], R[high]
        i = low
        j = high - 1
        while True:
            while i < high and LT(GET(A, R[i]), pivot):
                i += 1
            while j >= low and LT(pivot, GET(A, R[j])):
                j -= 1
            if i >= j:
                break
            R[i], R[j] = R[j], R[i]
            i += 1
            j -= 1
        # Put the pivot back in its final place (all items before `i`
        # are smaller than the pivot, all items at/after `i` are larger)
        R[i], R[high] = R[high], R[i]
        return i

    @wrap
    def partition3(A, low, high):
        """
        Three-way partition [low, high) around a chosen pivot.
        A tuple (lt, gt) is returned such that:
            - all elements in [low, lt) are < pivot
            - all elements in [lt, gt] are == pivot
            - all elements in (gt, high] are > pivot
        """
        mid = (low + high) >> 1
        # median of three {low, middle, high}
        if LT(A[mid], A[low]):
            A[low], A[mid] = A[mid], A[low]
        if LT(A[high], A[mid]):
            A[high], A[mid] = A[mid], A[high]
        if LT(A[mid], A[low]):
            A[low], A[mid] = A[mid], A[low]
        pivot = A[mid]

        A[low], A[mid] = A[mid], A[low]
        lt = low
        gt = high
        i = low + 1
        while i <= gt:
            if LT(A[i], pivot):
                A[lt], A[i] = A[i], A[lt]
                lt += 1
                i += 1
            elif LT(pivot, A[i]):
                A[gt], A[i] = A[i], A[gt]
                gt -= 1
            else:
                i += 1
        return lt, gt

    @wrap
    def run_quicksort1(A):
        R = make_res(A)

        if len(A) < 2:
            return R

        stack = [Partition(zero, zero)] * MAX_STACK
        stack[0] = Partition(zero, len(A) - 1)
        n = 1

        while n > 0:
            n -= 1
            low, high = stack[n]
            # Partition until it becomes more efficient to do an insertion sort
            while high - low >= SMALL_QUICKSORT:
                assert n < MAX_STACK
                i = partition(A, R, low, high)
                # Push largest partition on the stack
                if high - i > i - low:
                    # Right is larger
                    if high > i:
                        stack[n] = Partition(i + 1, high)
                        n += 1
                    high = i - 1
                else:
                    if i > low:
                        stack[n] = Partition(low, i - 1)
                        n += 1
                    low = i + 1

            insertion_sort(A, R, low, high)

        return R

    if is_np_array:
        @wrap
        def run_quicksort(A):
            if A.ndim == 1:
                return run_quicksort1(A)
            else:
                for idx in np.ndindex(A.shape[:-1]):
                    run_quicksort1(A[idx])
                return A
    else:
        @wrap
        def run_quicksort(A):
            return run_quicksort1(A)

    # Unused quicksort implementation based on 3-way partitioning; the
    # partitioning scheme turns out exhibiting bad behaviour on sorted arrays.
    @wrap
    def _run_quicksort(A):
        stack = [Partition(zero, zero)] * 100
        stack[0] = Partition(zero, len(A) - 1)
        n = 1

        while n > 0:
            n -= 1
            low, high = stack[n]
            # Partition until it becomes more efficient to do an insertion sort
            while high - low >= SMALL_QUICKSORT:
                assert n < MAX_STACK
                l, r = partition3(A, low, high)
                # One trivial (empty) partition => iterate on the other
                if r == high:
                    high = l - 1
                elif l == low:
                    low = r + 1
                # Push largest partition on the stack
                elif high - r > l - low:
                    # Right is larger
                    stack[n] = Partition(r + 1, high)
                    n += 1
                    high = l - 1
                else:
                    stack[n] = Partition(low, l - 1)
                    n += 1
                    low = r + 1

            insertion_sort(A, low, high)


    return QuicksortImplementation(wrap,
                                   partition, partition3, insertion_sort,
                                   run_quicksort)


def make_py_quicksort(*args, **kwargs):
    return make_quicksort_impl((lambda f: f), *args, **kwargs)

def make_jit_quicksort(*args, **kwargs):
    from numba.core.extending import register_jitable
    return make_quicksort_impl((lambda f: register_jitable(f)),
                               *args, **kwargs)



from numba import vectorize


def add(a, b):
    """An addition"""
    return a + b


def equals(a, b):
    return a == b


def mul(a, b):
    """A multiplication"""
    return a * b


def guadd(a, b, c):
    """A generalized addition"""
    x, y = c.shape
    for i in range(x):
        for j in range(y):
            c[i, j] = a[i, j] + b[i, j]


@vectorize(nopython=True)
def inner(a, b):
    return a + b


@vectorize(["int64(int64, int64)"], nopython=True)
def inner_explicit(a, b):
    return a + b


def outer(a, b):
    return inner(a, b)


def outer_explicit(a, b):
    return inner_explicit(a, b)


class Dummy:
    pass


def guadd_obj(a, b, c):
    Dummy()  # to force object mode
    x, y = c.shape
    for i in range(x):
        for j in range(y):
            c[i, j] = a[i, j] + b[i, j]


def guadd_scalar_obj(a, b, c):
    Dummy()  # to force object mode
    x, y = c.shape
    for i in range(x):
        for j in range(y):
            c[i, j] = a[i, j] + b


class MyException(Exception):
    pass


def guerror(a, b, c):
    raise MyException



import itertools

import numpy as np

from numba import jit, njit, typeof
from numba.core import types
from numba.tests.support import TestCase, MemoryLeakMixin
import unittest


def array_iter(arr):
    total = 0
    for i, v in enumerate(arr):
        total += i * v
    return total

def array_iter_items(arr):
    return list(iter(arr))

def array_view_iter(arr, idx):
    total = 0
    for i, v in enumerate(arr[idx]):
        total += i * v
    return total

def array_flat(arr, out):
    for i, v in enumerate(arr.flat):
        out[i] = v

def array_flat_getitem(arr, ind):
    return arr.flat[ind]

def array_flat_setitem(arr, ind, val):
    arr.flat[ind] = val

def array_flat_sum(arr):
    s = 0
    for i, v in enumerate(arr.flat):
        s = s + (i + 1) * v
    return s

def array_flat_len(arr):
    return len(arr.flat)

def array_ndenumerate_zero_dim(arr):
    for i, v in np.ndenumerate(arr):
        return (i, v)
    return None

def array_ndenumerate_sum(arr):
    s = 0
    for (i, j), v in np.ndenumerate(arr):
        s = s + (i + 1) * (j + 1) * v
    return s

def np_ndindex_empty():
    s = 0
    for ind in np.ndindex(()):
        s += s + len(ind) + 1
    return s

def np_ndindex(x, y):
    s = 0
    n = 0
    for i, j in np.ndindex(x, y):
        s = s + (i + 1) * (j + 1)
    return s

def np_ndindex_array(arr):
    s = 0
    n = 0
    for indices in np.ndindex(arr.shape):
        for i, j in enumerate(indices):
            s = s + (i + 1) * (j + 1)
    return s

def np_nditer1(a):
    res = []
    for u in np.nditer(a):
        res.append(u.item())
    return res

def np_nditer2(a, b):
    res = []
    for u, v in np.nditer((a, b)):
        res.append((u.item(), v.item()))
    return res

def np_nditer3(a, b, c):
    res = []
    for u, v, w in np.nditer((a, b, c)):
        res.append((u.item(), v.item(), w.item()))
    return res

def iter_next(arr):
    it = iter(arr)
    it2 = iter(arr)
    return next(it), next(it), next(it2)


#
# Test premature free (see issue #2112).
# The following test allocates an array ``x`` inside the body.
# The compiler will put a ``del x`` right after the last use of ``x``,
# which is right after the creation of the array iterator and
# before the loop is entered.  If the iterator does not incref the array,
# the iterator will be reading garbage data of free'ed memory.
#

def array_flat_premature_free(size):
    x = np.arange(size)
    res = np.zeros_like(x, dtype=np.intp)
    for i, v in enumerate(x.flat):
        res[i] = v
    return res

def array_ndenumerate_premature_free(size):
    x = np.arange(size)
    res = np.zeros_like(x, dtype=np.intp)
    for i, v in np.ndenumerate(x):
        res[i] = v
    return res


class TestArrayIterators(MemoryLeakMixin, TestCase):
    """
    Test array.flat, np.ndenumerate(), etc.
    """

    def setUp(self):
        super(TestArrayIterators, self).setUp()

    def check_array_iter_1d(self, arr):
        pyfunc = array_iter
        cfunc = njit((typeof(arr),))(pyfunc)
        expected = pyfunc(arr)
        self.assertPreciseEqual(cfunc(arr), expected)

    def check_array_iter_items(self, arr):
        pyfunc = array_iter_items
        cfunc = njit((typeof(arr),))(pyfunc)
        expected = pyfunc(arr)
        self.assertPreciseEqual(cfunc(arr), expected)

    def check_array_view_iter(self, arr, index):
        pyfunc = array_view_iter
        cfunc = njit((typeof(arr), typeof(index),))(pyfunc)
        expected = pyfunc(arr, index)
        self.assertPreciseEqual(cfunc(arr, index), expected)

    def check_array_flat(self, arr, arrty=None):
        out = np.zeros(arr.size, dtype=arr.dtype)
        nb_out = out.copy()
        if arrty is None:
            arrty = typeof(arr)

        cfunc = njit((arrty, typeof(out),))(array_flat)

        array_flat(arr, out)
        cfunc(arr, nb_out)

        self.assertPreciseEqual(out, nb_out)

    def check_array_unary(self, arr, arrty, func):
        cfunc = njit((arrty,))(func)
        self.assertPreciseEqual(cfunc(arr), func(arr))

    def check_array_ndenumerate_sum(self, arr, arrty):
        self.check_array_unary(arr, arrty, array_ndenumerate_sum)

    def test_array_iter(self):
        # Test iterating over arrays
        arr = np.arange(6)
        self.check_array_iter_1d(arr)
        self.check_array_iter_items(arr)
        arr = arr[::2]
        self.assertFalse(arr.flags.c_contiguous)
        self.assertFalse(arr.flags.f_contiguous)
        self.check_array_iter_1d(arr)
        self.check_array_iter_items(arr)
        arr = np.bool_([1, 0, 0, 1])
        self.check_array_iter_1d(arr)
        self.check_array_iter_items(arr)
        arr = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        self.check_array_iter_items(arr)
        self.check_array_iter_items(arr.T)

    def test_array_iter_yielded_order(self):
        # See issue #5692
        @jit(nopython=True)
        def foo(arr):
            t = []
            for y1 in arr:
                for y2 in y1:
                    t.append(y2.ravel())
            return t

        # 'F' ordered
        arr = np.arange(24).reshape((2, 3, 4), order='F')
        expected = foo.py_func(arr)
        got = foo(arr)
        self.assertPreciseEqual(expected, got)

        # 'A' ordered, outer strided
        arr = np.arange(64).reshape((4, 8, 2), order='F')[::2, :, :]
        expected = foo.py_func(arr)
        got = foo(arr)
        self.assertPreciseEqual(expected, got)

        # 'A' ordered, middle strided
        arr = np.arange(64).reshape((4, 8, 2), order='F')[:, ::2, :]
        expected = foo.py_func(arr)
        got = foo(arr)
        self.assertPreciseEqual(expected, got)

        # 'A' ordered, inner strided
        arr = np.arange(64).reshape((4, 8, 2), order='F')[:, :, ::2]
        expected = foo.py_func(arr)
        got = foo(arr)
        self.assertPreciseEqual(expected, got)

        @jit(nopython=True)
        def flag_check(arr):
            out = []
            for sub in arr:
                out.append((sub, sub.flags.c_contiguous,
                            sub.flags.f_contiguous))
            return out

        arr = np.arange(10).reshape((2, 5), order='F')
        expected = flag_check.py_func(arr)
        got = flag_check(arr)

        self.assertEqual(len(expected), len(got))
        ex_arr, e_flag_c, e_flag_f = expected[0]
        go_arr, g_flag_c, g_flag_f = got[0]
        np.testing.assert_allclose(ex_arr, go_arr)
        self.assertEqual(e_flag_c, g_flag_c)
        self.assertEqual(e_flag_f, g_flag_f)

    def test_array_view_iter(self):
        # Test iterating over a 1d view over a 2d array
        arr = np.arange(12).reshape((3, 4))
        self.check_array_view_iter(arr, 1)
        self.check_array_view_iter(arr.T, 1)
        arr = arr[::2]
        self.check_array_view_iter(arr, 1)
        arr = np.bool_([1, 0, 0, 1]).reshape((2, 2))
        self.check_array_view_iter(arr, 1)

    def test_array_flat_3d(self):
        arr = np.arange(24).reshape(4, 2, 3)

        arrty = typeof(arr)
        self.assertEqual(arrty.ndim, 3)
        self.assertEqual(arrty.layout, 'C')
        self.assertTrue(arr.flags.c_contiguous)
        # Test with C-contiguous array
        self.check_array_flat(arr)
        # Test with Fortran-contiguous array
        arr = arr.transpose()
        self.assertFalse(arr.flags.c_contiguous)
        self.assertTrue(arr.flags.f_contiguous)
        self.assertEqual(typeof(arr).layout, 'F')
        self.check_array_flat(arr)
        # Test with non-contiguous array
        arr = arr[::2]
        self.assertFalse(arr.flags.c_contiguous)
        self.assertFalse(arr.flags.f_contiguous)
        self.assertEqual(typeof(arr).layout, 'A')
        self.check_array_flat(arr)
        # Boolean array
        arr = np.bool_([1, 0, 0, 1] * 2).reshape((2, 2, 2))
        self.check_array_flat(arr)

    def test_array_flat_empty(self):
        # Test .flat with various shapes of empty arrays, contiguous
        # and non-contiguous (see issue #846).

        # Define a local checking function, Numba's `typeof` ends up aliasing
        # 0d C and F ordered arrays, so the check needs to go via the compile
        # result entry point to bypass type checking.
        def check(arr, arrty):
            cfunc = njit((arrty,))(array_flat_sum)
            cres = cfunc.overloads[(arrty,)]
            got = cres.entry_point(arr)
            expected = cfunc.py_func(arr)
            self.assertPreciseEqual(expected, got)

        arr = np.zeros(0, dtype=np.int32)
        arr = arr.reshape(0, 2)
        arrty = types.Array(types.int32, 2, layout='C')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='F')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='A')
        check(arr, arrty)
        arr = arr.reshape(2, 0)
        arrty = types.Array(types.int32, 2, layout='C')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='F')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='A')
        check(arr, arrty)

    def test_array_flat_getitem(self):
        # Test indexing of array.flat object
        pyfunc = array_flat_getitem
        cfunc = njit(pyfunc)
        def check(arr, ind):
            expected = pyfunc(arr, ind)
            self.assertEqual(cfunc(arr, ind), expected)

        arr = np.arange(24).reshape(4, 2, 3)
        for i in range(arr.size):
            check(arr, i)
        arr = arr.T
        for i in range(arr.size):
            check(arr, i)
        arr = arr[::2]
        for i in range(arr.size):
            check(arr, i)
        arr = np.array([42]).reshape(())
        for i in range(arr.size):
            check(arr, i)
        # Boolean array
        arr = np.bool_([1, 0, 0, 1])
        for i in range(arr.size):
            check(arr, i)
        arr = arr[::2]
        for i in range(arr.size):
            check(arr, i)

    def test_array_flat_setitem(self):
        # Test indexing of array.flat object
        pyfunc = array_flat_setitem
        cfunc = njit(pyfunc)
        def check(arr, ind):
            # Use np.copy() to keep the layout
            expected = np.copy(arr)
            got = np.copy(arr)
            pyfunc(expected, ind, 123)
            cfunc(got, ind, 123)
            self.assertPreciseEqual(got, expected)

        arr = np.arange(24).reshape(4, 2, 3)
        for i in range(arr.size):
            check(arr, i)
        arr = arr.T
        for i in range(arr.size):
            check(arr, i)
        arr = arr[::2]
        for i in range(arr.size):
            check(arr, i)
        arr = np.array([42]).reshape(())
        for i in range(arr.size):
            check(arr, i)
        # Boolean array
        arr = np.bool_([1, 0, 0, 1])
        for i in range(arr.size):
            check(arr, i)
        arr = arr[::2]
        for i in range(arr.size):
            check(arr, i)

    def test_array_flat_len(self):
        # Test len(array.flat)
        pyfunc = array_flat_len
        cfunc = njit(array_flat_len)
        def check(arr):
            expected = pyfunc(arr)
            self.assertPreciseEqual(cfunc(arr), expected)

        arr = np.arange(24).reshape(4, 2, 3)
        check(arr)
        arr = arr.T
        check(arr)
        arr = arr[::2]
        check(arr)
        arr = np.array([42]).reshape(())
        check(arr)

    def test_array_flat_premature_free(self):
        cfunc = njit((types.intp,))(array_flat_premature_free)
        expect = array_flat_premature_free(6)
        got = cfunc(6)
        self.assertTrue(got.sum())
        self.assertPreciseEqual(expect, got)

    def test_array_ndenumerate_2d(self):
        arr = np.arange(12).reshape(4, 3)
        arrty = typeof(arr)
        self.assertEqual(arrty.ndim, 2)
        self.assertEqual(arrty.layout, 'C')
        self.assertTrue(arr.flags.c_contiguous)
        # Test with C-contiguous array
        self.check_array_ndenumerate_sum(arr, arrty)
        # Test with Fortran-contiguous array
        arr = arr.transpose()
        self.assertFalse(arr.flags.c_contiguous)
        self.assertTrue(arr.flags.f_contiguous)
        arrty = typeof(arr)
        self.assertEqual(arrty.layout, 'F')
        self.check_array_ndenumerate_sum(arr, arrty)
        # Test with non-contiguous array
        arr = arr[::2]
        self.assertFalse(arr.flags.c_contiguous)
        self.assertFalse(arr.flags.f_contiguous)
        arrty = typeof(arr)
        self.assertEqual(arrty.layout, 'A')
        self.check_array_ndenumerate_sum(arr, arrty)
        # Boolean array
        arr = np.bool_([1, 0, 0, 1]).reshape((2, 2))
        self.check_array_ndenumerate_sum(arr, typeof(arr))

    def test_array_ndenumerate_empty(self):
        # Define a local checking function, Numba's `typeof` ends up aliasing
        # 0d C and F ordered arrays, so the check needs to go via the compile
        # result entry point to bypass type checking.
        def check(arr, arrty):
            cfunc = njit((arrty,))(array_ndenumerate_sum)
            cres = cfunc.overloads[(arrty,)]
            got = cres.entry_point(arr)
            expected = cfunc.py_func(arr)
            np.testing.assert_allclose(expected, got)

        arr = np.zeros(0, dtype=np.int32)
        arr = arr.reshape(0, 2)
        arrty = types.Array(types.int32, 2, layout='C')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='F')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='A')
        check(arr, arrty)
        arr = arr.reshape(2, 0)
        arrty = types.Array(types.int32, 2, layout='C')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='F')
        check(arr, arrty)
        arrty = types.Array(types.int32, 2, layout='A')
        check(arr, arrty)

    def test_array_ndenumerate_premature_free(self):
        cfunc = njit((types.intp,))(array_ndenumerate_premature_free)
        expect = array_ndenumerate_premature_free(6)
        got = cfunc(6)
        self.assertTrue(got.sum())
        self.assertPreciseEqual(expect, got)

    def test_array_ndenumerate_zero_dim(self):
        func = array_ndenumerate_zero_dim
        cfunc = njit(func)
        arr = np.array(97)
        expected = func(arr)
        got = cfunc(arr)
        self.assertPreciseEqual(got, expected)
        self.assertEqual(expected[0], ())
        self.assertEqual(expected[1], 97)

    def test_np_ndindex(self):
        func = np_ndindex
        cfunc = njit((types.int32, types.int32,))(func)
        self.assertPreciseEqual(cfunc(3, 4), func(3, 4))
        self.assertPreciseEqual(cfunc(3, 0), func(3, 0))
        self.assertPreciseEqual(cfunc(0, 3), func(0, 3))
        self.assertPreciseEqual(cfunc(0, 0), func(0, 0))

    def test_np_ndindex_array(self):
        func = np_ndindex_array
        arr = np.arange(12, dtype=np.int32) + 10
        self.check_array_unary(arr, typeof(arr), func)
        arr = arr.reshape((4, 3))
        self.check_array_unary(arr, typeof(arr), func)
        arr = arr.reshape((2, 2, 3))
        self.check_array_unary(arr, typeof(arr), func)

    def test_np_ndindex_empty(self):
        func = np_ndindex_empty
        cfunc = njit((),)(func)
        self.assertPreciseEqual(cfunc(), func())

    def test_iter_next(self):
        # This also checks memory management with iter() and next()
        func = iter_next
        arr = np.arange(12, dtype=np.int32) + 10
        self.check_array_unary(arr, typeof(arr), func)


class TestNdIter(MemoryLeakMixin, TestCase):
    """
    Test np.nditer()
    """

    def inputs(self):
        # All those inputs are compatible with a (3, 4) main shape

        # scalars
        yield np.float32(100)

        # 0-d arrays
        yield np.array(102, dtype=np.int16)

        # 1-d arrays
        yield np.arange(4).astype(np.complex64)
        yield np.arange(8)[::2]

        # 2-d arrays
        a = np.arange(12).reshape((3, 4))
        yield a
        yield a.copy(order='F')
        a = np.arange(24).reshape((6, 4))[::2]
        yield a

    def basic_inputs(self):
        yield np.arange(4).astype(np.complex64)
        yield np.arange(8)[::2]
        a = np.arange(12).reshape((3, 4))
        yield a
        yield a.copy(order='F')

    def check_result(self, got, expected):
        self.assertEqual(set(got), set(expected), (got, expected))

    def test_nditer1(self):
        pyfunc = np_nditer1
        cfunc = jit(nopython=True)(pyfunc)
        for a in self.inputs():
            expected = pyfunc(a)
            got = cfunc(a)
            self.check_result(got, expected)

    def test_nditer2(self):
        pyfunc = np_nditer2
        cfunc = jit(nopython=True)(pyfunc)
        for a, b in itertools.product(self.inputs(), self.inputs()):
            expected = pyfunc(a, b)
            got = cfunc(a, b)
            self.check_result(got, expected)

    def test_nditer3(self):
        pyfunc = np_nditer3
        cfunc = jit(nopython=True)(pyfunc)
        # Use a restricted set of inputs, to shorten test time
        inputs = self.basic_inputs
        for a, b, c in itertools.product(inputs(), inputs(), inputs()):
            expected = pyfunc(a, b, c)
            got = cfunc(a, b, c)
            self.check_result(got, expected)

    def test_errors(self):
        # Incompatible shapes
        pyfunc = np_nditer2
        cfunc = jit(nopython=True)(pyfunc)

        self.disable_leak_check()

        def check_incompatible(a, b):
            with self.assertRaises(ValueError) as raises:
                cfunc(a, b)
            self.assertIn("operands could not be broadcast together",
                          str(raises.exception))

        check_incompatible(np.arange(2), np.arange(3))
        a = np.arange(12).reshape((3, 4))
        b = np.arange(3)
        check_incompatible(a, b)


if __name__ == '__main__':
    unittest.main()
